import re
from datetime import datetime, timedelta, timezone
from typing import Union


class TimeExpressionNotRecognized(Exception):
    def __init__(self, time_str):
        self.message = f"Time expression could not be parsed: {time_str}"
        super(TimeExpressionNotRecognized, self).__init__(self.message)


class ErrorParsingTime(Exception):
    def __init__(self, message):
        self.message = message
        super(ErrorParsingTime, self).__init__(message)


def parse_hour_minute(input_str):
    split_str = input_str.split(":")
    hour = int(split_str[0])
    minute = int(split_str[1])
    return hour, minute


def parse_day_month_DD_MM(input_str):
    split_str = input_str.split(".")
    day = int(split_str[0])
    month = int(split_str[1])
    return day, month


def parse_day_month_MM_DD(input_str):
    split_str = input_str.split("/")
    month = int(split_str[0])
    day = int(split_str[1])
    return day, month


def parse_day_month_DD_MM_YYorYYYY(input_str):
    split_str = input_str.split(".")
    day = int(split_str[0])
    month = int(split_str[1])
    year = split_str[2]
    if len(year) == 2:
        year = int("20" + year)
    else:
        year = int(year)
    return day, month, year


def add_day_if_past(dt: datetime) -> datetime:
    """This function will add a day to the datetime object 'dt' when 'dt' is in the past"""
    dt_now = datetime.now()
    if dt < dt_now:
        return dt + timedelta(days=1)
    else:
        return dt


def parse_datetime(datetime_str: str):
    try:
        if re.match(r"([0-9]{1,2}h)", datetime_str, re.IGNORECASE):
            """e.g. 1h / 12h"""
            return datetime.now() + timedelta(hours=int(datetime_str[:-1]))

        if datetime_str == "morning":
            dt = datetime.now()
            return add_day_if_past(
                dt.replace(hour=7, minute=0, second=0, microsecond=0)
            )

        if datetime_str == "tomorrow":
            dt = datetime.now()
            return dt.replace(hour=7, minute=0, second=0, microsecond=0) + timedelta(
                days=1
            )

        if datetime_str == "evening":
            dt = datetime.now()
            return add_day_if_past(
                dt.replace(hour=18, minute=0, second=0, microsecond=0)
            )

        if re.match(r"([0-9]{1,2}:[0-9]{2} (am|pm))", datetime_str, re.IGNORECASE):
            """e.g. 5:30 pm"""
            split_str = datetime_str.split(" ")
            hour, minute = parse_hour_minute(split_str[0])

            if split_str[1] == "am":
                if hour == 12:
                    hour = 0
            else:
                hour = hour + 12

            return add_day_if_past(
                datetime.now().replace(
                    hour=hour, minute=minute, second=0, microsecond=0
                )
            )

        if re.match(r"([0-9]{1,2}:[0-9]{2})", datetime_str, re.IGNORECASE):
            """e.g. 17:00"""
            hour, minute = parse_hour_minute(datetime_str)
            return add_day_if_past(
                datetime.now().replace(
                    hour=hour, minute=minute, second=0, microsecond=0
                )
            )

        if re.match(
            r"([0-9]{1,2}\.[0-9]{1,2}\.(([0-9]{4})|([0-9]{2})))$",
            datetime_str,
            re.IGNORECASE,
        ):
            """e.g. 17.01.20 or 22.12.2020"""
            day, month, year = parse_day_month_DD_MM_YYorYYYY(datetime_str)
            return datetime.now().replace(
                year=year,
                day=day,
                month=month,
                hour=7,
                minute=0,
                second=0,
                microsecond=0,
            )

        if re.match(
            r"([0-9]{1,2}\.[0-9]{1,2}\. [0-9]{1,2}:[0-9]{2})",
            datetime_str,
            re.IGNORECASE,
        ):
            """e.g. 17.01. 17:00"""
            split_str = datetime_str.split(" ")
            day, month = parse_day_month_DD_MM(split_str[0])
            hour, minute = parse_hour_minute(split_str[1])
            return datetime.now().replace(
                day=day, month=month, hour=hour, minute=minute, second=0, microsecond=0
            )

        # if re.match(
        #     r"([0-9]{1,2}/[0-9]{1,2} [0-9]{1,2}:[0-9]{2} (am|pm))",
        #     datetime_str,
        #     re.IGNORECASE,
        # ):
        #     """ e.g. 01/17 5:00 pm """
        #     split_str = datetime_str.split(" ")
        #     day, month = parse_day_month_MM_DD(split_str[0])
        #     hour, minute = parse_hour_minute(split_str[1])
        #     return datetime.now().replace(
        #         day=day, month=month, hour=hour, minute=minute, second=0, microsecond=0
        #     )

    except ValueError as e:
        raise ErrorParsingTime(str(e))

    raise TimeExpressionNotRecognized(datetime_str)


def datetime_to_api_timestamp(dt: datetime):
    utc_dt = dt.astimezone(timezone.utc)
    timestamp_str = utc_dt.strftime("%Y-%m-%dT%H:%M:%S")

    api_dt = {"dateTime": timestamp_str, "timeZone": "UTC"}
    return api_dt


def api_timestamp_to_datetime(api_dt: Union[str, dict]):
    """Converts the datetime string returned by the API to python datetime object"""

    """
    Somehow this string is formatted with 7 digits for 'microsecond' resolution, so crop the last digit (and trailing Z)
    The cropped string will be written into api_dt_str_mod
    """
    api_dt_str_mod = None

    if isinstance(api_dt, str):
        api_dt_str_mod = api_dt[:-2]
    elif isinstance(api_dt, dict):
        api_dt_str_mod = api_dt["dateTime"][:-2]
    else:
        raise

    dt = datetime.strptime(api_dt_str_mod, "%Y-%m-%dT%H:%M:%S.%f")
    dt = utc_to_local(dt)
    return dt


def utc_to_local(_dt):
    return _dt.replace(tzinfo=timezone.utc).astimezone(tz=None)



#!/usr/bin/env python
from asyncio import tasks
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import (
    KeyBindings,
    ConditionalKeyBindings,
    merge_key_bindings,
)
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    HSplit,
    VSplit,
    Window,
    WindowAlign,
    DynamicContainer,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import TextArea

import todocli.graphapi.todo_api as todo_api

# Colors
COLOR_FOLDER = "bg:#006699"
COLOR_TASK = "bg:#006699"

global_focus_index_folder = 0
global_focus_index_task = 0
global_focus_on_folder = True

# This flag is used to direct user input to confirmation prompt
global_waiting_confirmation = False

# Global data structures for ui
global_folder2id = {}
global_folders = []
global_task2id = {}
global_tasks_ui = []
global_tasks = []


def load_folders():
    """
    Load all folders
    """
    global left_window
    global global_focus_index_folder
    global global_folder2id
    global global_folders

    # Reset all folder data structures
    global_focus_index_folder = 0
    global_folder2id.clear()
    global_folders.clear()

    # Retrieve folder data
    lists = todo_api.query_lists()
    for idx, _list in enumerate(lists):
        global_folder2id[idx] = _list.id
        global_folders.append(_list.display_name)

    # Layout interface
    left_window.children = [
        Window(FormattedTextControl(f), height=1, width=50) for f in global_folders
    ]

    # Highlight first folder
    left_window.children[0].style = COLOR_FOLDER


left_window = HSplit([Window()])

# Get folders to load on screen
load_folders()

right_window = HSplit([Window()])
body = VSplit(
    [
        left_window,
        Window(width=1, char="|", style="class:line"),
        right_window,
    ]
)


def get_titlebar_text():
    return [("class:title", " tod0 ")]


prompt_window = Window(
    height=1, content=FormattedTextControl(""), align=WindowAlign.LEFT
)

root_container = HSplit(
    [
        # The titlebar.
        Window(
            height=1,
            content=FormattedTextControl(get_titlebar_text),
            align=WindowAlign.CENTER,
        ),
        # Horizontal separator.
        Window(height=1, char="-", style="class:line"),
        # The 'body', like defined above.
        body,
        Window(height=1, char=".", style="class:line"),
        DynamicContainer(lambda: prompt_window),
    ]
)


# Key bindings
kb = KeyBindings()
kb_exit = KeyBindings()
kb_escape = KeyBindings()


@kb_exit.add("c-c", eager=True)
@kb_exit.add("c-q", eager=True)
def _(event):
    """
    Pressing Ctrl-Q or Ctrl-C will exit the user interface.
    """
    event.app.exit()


@kb.add("?")
def _(event):
    """
    Pressing Shift-? will display help toolbar.
    """
    global global_waiting_confirmation
    global prompt_window

    global_waiting_confirmation = True

    input_field = TextArea(
        height=1,
        prompt="[UP: j] [DOWN: k] [SELECT: l] [BACK: h] [CREATE: n] [MARK COMPLETE: c] [EXIT HELP: ESC]",
        style="class:output-field",
        multiline=False,
        wrap_lines=False,
    )

    prompt_window = input_field
    event.app.layout.focus(input_field)


@kb.add("j")
def _(event):
    """
    Move selection down 1
    """

    if global_focus_on_folder:
        global global_focus_index_folder
        folder_window = left_window.children
        folder_window[global_focus_index_folder].style = ""
        global_focus_index_folder = (global_focus_index_folder + 1) % len(
            global_folders
        )
        folder_window[global_focus_index_folder].style = COLOR_FOLDER
    else:
        global global_focus_index_task
        if global_tasks_ui:
            global_tasks_ui[global_focus_index_task].style = ""
            global_focus_index_task = (global_focus_index_task + 1) % len(
                global_tasks_ui
            )
            global_tasks_ui[global_focus_index_task].style = COLOR_TASK


@kb.add("k")
def _(event):
    """
    Move selection up 1
    """

    if global_focus_on_folder:
        global global_focus_index_folder
        folder_window = left_window.children
        folder_window[global_focus_index_folder].style = ""
        global_focus_index_folder = (global_focus_index_folder - 1) % len(
            global_folders
        )
        folder_window[global_focus_index_folder].style = COLOR_FOLDER
    else:
        global global_focus_index_task
        if global_tasks_ui:
            global_tasks_ui[global_focus_index_task].style = ""
            global_focus_index_task = (global_focus_index_task - 1) % len(
                global_tasks_ui
            )
            global_tasks_ui[global_focus_index_task].style = COLOR_TASK


@kb.add("l")
def _(event):
    """
    Select currently focused folder
    """
    if global_focus_on_folder:
        load_tasks()


@kb.add("h")
def _(event):
    """
    Go back to folder scroll mode
    """
    global global_focus_index_task
    global global_focus_on_folder
    global right_window

    if global_tasks_ui:
        global_tasks_ui[global_focus_index_task].style = ""
    right_window.children = [Window()]
    global_focus_on_folder = True


@kb.add("c")
def _(event):
    """
    Mark task as complete
    """

    # Only receive input on task view mode
    if global_focus_on_folder or (not global_focus_on_folder and not global_tasks_ui):
        return

    global global_waiting_confirmation
    global prompt_window
    global global_tasks

    global_waiting_confirmation = True

    input_field = TextArea(
        height=1,
        prompt="Mark task as complete? <y> to confirm. ",
        style="class:input-field",
        multiline=False,
        wrap_lines=False,
    )

    # Confirmation of commands
    def confirm(buff):
        global global_waiting_confirmation
        global prompt_window
        user_input = input_field.text
        if user_input == "y":
            # Mark task as complete
            todo_api.complete_task(
                global_folders[global_focus_index_folder],
                global_tasks[global_focus_index_task].title,
            )
            load_tasks()

        # Return to normal state
        global_waiting_confirmation = False
        prompt_window = Window()

    input_field.accept_handler = confirm

    prompt_window = input_field
    event.app.layout.focus(input_field)


@kb.add("n")
def _(event):
    """
    Create new task/folder
    """
    global global_waiting_confirmation
    global prompt_window

    global_waiting_confirmation = True

    # Check if we are creating new task or folder
    if global_focus_on_folder:
        # We are creating a new folder
        input_field = TextArea(
            height=1,
            prompt="New folder: ",
            style="class:input-field",
            multiline=False,
            wrap_lines=False,
        )

        # Get new folder name
        def get_name(buff):
            global global_waiting_confirmation
            global prompt_window
            global left_window

            user_input = input_field.text
            if user_input:
                # Create new folder
                todo_api.create_list(user_input)
                # Refresh folders
                load_folders()

            # Return to normal state
            global_waiting_confirmation = False
            prompt_window = Window()

    else:
        # We are creating a new task
        input_field = TextArea(
            height=1,
            prompt="New task: ",
            style="class:input-field",
            multiline=False,
            wrap_lines=False,
        )

        # Get new task name
        def get_name(buff):
            global global_waiting_confirmation
            global prompt_window
            user_input = input_field.text

            if user_input:
                # Create new task
                todo_api.create_task(
                    user_input, global_folders[global_focus_index_folder]
                )
                # Refresh tasks
                load_tasks()

            # Return to normal state
            global_waiting_confirmation = False
            prompt_window = Window()

    input_field.accept_handler = get_name

    prompt_window = input_field
    event.app.layout.focus(input_field)


@kb_escape.add("escape", eager=True)
def _(event):
    """
    Escape prompt
    """
    global global_waiting_confirmation
    global prompt_window

    # Return to normal state
    global_waiting_confirmation = False
    prompt_window = Window()


@Condition
def is_not_waiting_for_confirmation():
    "Enable key bindings when not waiting for confirmation"
    return not global_waiting_confirmation


kb = ConditionalKeyBindings(kb, is_not_waiting_for_confirmation)


def load_tasks():
    """
    Load tasks of currently focused folder
    """

    global global_focus_on_folder
    global global_focus_index_task
    global global_tasks_ui
    global global_task2id
    global global_tasks

    folder_name = global_folders[global_focus_index_folder]
    global_tasks = todo_api.query_tasks(folder_name, num_tasks=100)

    global_tasks_ui = []
    for idx, t in enumerate(global_tasks):
        id_ = t.id
        title = t.title
        global_tasks_ui.append(Window(FormattedTextControl(title), height=1))
        global_task2id[idx] = id_

    # Add empty container if task list is empty
    if not global_tasks_ui:
        right_window.children = [Window(FormattedTextControl("-- No Tasks --"))]
    else:
        right_window.children = global_tasks_ui
        global_focus_index_task = 0
        global_tasks_ui[global_focus_index_task].style = COLOR_TASK
    global_focus_on_folder = False


# Creating an `Application` instance
# ----------------------------------
application = Application(
    layout=Layout(root_container),
    key_bindings=merge_key_bindings([kb, kb_exit, kb_escape]),
    mouse_support=False,
    full_screen=False,
)


# Run the application
# -------------------
def run():
    application.run()


if __name__ == "__main__":
    run()



# Oauth settings
import os
import pickle
import time

import yaml
from requests_oauthlib import OAuth2Session

settings = {
    "redirect": "https://localhost/login/authorized",
    "scopes": "openid offline_access tasks.readwrite",
    "authority": "https://login.microsoftonline.com/common",
    "authorize_endpoint": "/oauth2/v2.0/authorize",
    "token_endpoint": "/oauth2/v2.0/token",
}

# Code taken from https://docs.microsoft.com/en-us/graph/tutorials/python?tutorial-step=3

# This is necessary because Azure does not guarantee
# to return scopes in the same case and order as requested
os.environ["OAUTHLIB_RELAX_TOKEN_SCOPE"] = "1"
os.environ["OAUTHLIB_IGNORE_SCOPE_CHANGE"] = "1"

redirect = settings["redirect"]
scope = settings["scopes"]

authorize_url = "{0}{1}".format(settings["authority"], settings["authorize_endpoint"])
token_url = "{0}{1}".format(settings["authority"], settings["token_endpoint"])

# User settings location
config_dir = "{}/.config/tod0".format(os.path.expanduser("~"))
if not os.path.isdir(config_dir):
    os.makedirs(config_dir)


def check_keys(keys):
    client_id = keys["client_id"]
    client_secret = keys["client_secret"]

    if client_id == "" or client_secret == "":
        print(
            "Please enter your client id and secret in {}".format(
                os.path.join(config_dir, "keys.yml")
            )
        )
        print(
            "Instructions to getting your API client id and secret can be found here:\n{}".format(
                "https://github.com/kiblee/tod0/blob/master/GET_KEY.md"
            )
        )
        exit()


# Check for api keys
keys_path = os.path.join(config_dir, "keys.yml")
if not os.path.isfile(keys_path):
    keys = {"client_id": "", "client_secret": ""}

    with open(keys_path, "w") as f:
        yaml.dump(keys, f)
    check_keys(keys)
else:
    # Load api keys
    with open(keys_path) as f:
        keys = yaml.load(f, yaml.SafeLoader)
        check_keys(keys)

client_id = keys["client_id"]
client_secret = keys["client_secret"]


def get_token():
    try:
        # Try to load token from local
        with open(os.path.join(config_dir, "token.pkl"), "rb") as f:
            token = pickle.load(f)

        token = refresh_token(token)

    except Exception:
        # Authorize user to get token
        outlook = OAuth2Session(client_id, scope=scope, redirect_uri=redirect)

        # Redirect  the user owner to the OAuth provider
        authorization_url, state = outlook.authorization_url(authorize_url)
        print("Please go here and authorize:\n", authorization_url)

        # Get the authorization verifier code from the callback url
        redirect_response = input("Paste the full redirect URL below:\n")

        # Fetch the access token
        token = outlook.fetch_token(
            token_url,
            client_secret=client_secret,
            authorization_response=redirect_response,
        )

    store_token(token)
    return token


def store_token(token):
    with open(os.path.join(config_dir, "token.pkl"), "wb") as f:
        pickle.dump(token, f)


def refresh_token(token):
    # Check expiration
    now = time.time()
    # Subtract 5 minutes from expiration to account for clock skew
    expire_time = token["expires_at"] - 300
    if now >= expire_time:
        # Refresh the token
        aad_auth = OAuth2Session(
            client_id, token=token, scope=scope, redirect_uri=redirect
        )

        refresh_params = {"client_id": client_id, "client_secret": client_secret}

        new_token = aad_auth.refresh_token(token_url, **refresh_params)
        return new_token

    # Token still valid, just return it
    return token


def get_oauth_session():
    token = get_token()
    return OAuth2Session(client_id, scope=scope, token=token)



import json

from todocli.graphapi.oauth import get_oauth_session


class RestRequest:
    def __init__(self, url):
        self.url = url

    @staticmethod
    def evaluateResult(o):
        if o.ok:
            return True
        else:
            o.raise_for_status()

    @staticmethod
    def parse_contents(response):
        return json.loads(response.content.decode())["value"]

    def parseResult(self, o):
        self.evaluateResult(o)
        return self.parse_contents(o)


class RestRequestWithBody(RestRequest):
    def __init__(self, url):
        super().__init__(url)
        self.body = {}

    def __setitem__(self, key, value):
        self.body[key] = value

    def addToRequestBody(self, tag, value):
        self.body[tag] = value


class RestRequestGet(RestRequest):
    def execute(self):
        outlook = get_oauth_session()
        o = outlook.get(self.url)
        return self.parseResult(o)


class RestRequestPost(RestRequestWithBody):
    def execute(self):
        outlook = get_oauth_session()
        o = outlook.post(self.url, json=self.body)
        return self.evaluateResult(o)


class RestRequestPatch(RestRequestWithBody):
    def execute(self):
        outlook = get_oauth_session()
        o = outlook.patch(self.url, json=self.body)
        return self.evaluateResult(o)


class RestRequestDelete(RestRequest):
    def execute(self):
        outlook = get_oauth_session()
        o = outlook.delete(self.url)
        return self.evaluateResult(o)



from enum import Enum


class TodoList:
    class WellKnownListName(Enum):
        none = "none"
        DefaultList = "defaultList"
        FlaggedEmails = "flaggedEmails"

    def __init__(self, query_result_list):
        self.id: str = query_result_list["id"]
        self.display_name: str = query_result_list["displayName"]
        self.is_owner = bool(query_result_list["isOwner"])
        self.is_shared = bool(query_result_list["isShared"])
        self.well_known_list_name = TodoList.WellKnownListName(
            query_result_list["wellknownListName"]
        )



import argparse
import shlex
import sys

import todocli.graphapi.todo_api as todo_api
from todocli.utils.datetime_util import (
    parse_datetime,
    TimeExpressionNotRecognized,
    ErrorParsingTime,
)
from todocli.error import eprint
from todocli.help_msg import help_msg


class InvalidTaskPath(Exception):
    def __init__(self, path):
        self.message = (
            "Invalid path: '{}', path can only contain one '/'. "
            "Please specify the task in format: '<list>/<task>'".format(path)
        )
        super(InvalidTaskPath, self).__init__(self.message)


def parse_task_path(task_path):
    if "/" in task_path:
        elems = task_path.split("/")
        if len(elems) > 2:
            raise InvalidTaskPath(task_path)
        return elems[0], elems[1]
    else:
        return "Tasks", task_path


def print_help():
    print(help_msg)


def print_list(item_list, print_line_nums):
    for i, x in enumerate(item_list):
        if print_line_nums:
            print("[{}] ".format(i), end="")
        print(x)


def ls(args):
    lists = todo_api.query_lists()
    lists_names = [l.display_name for l in lists]
    print_list(lists_names, args.display_linenums)


def lst(args):
    tasks = todo_api.query_tasks(args.list_name)
    tasks_titles = [x.title for x in tasks]
    print_list(tasks_titles, args.display_linenums)


def new(args):
    task_list, name = parse_task_path(args.task_name)

    reminder_date_time_str = args.reminder
    reminder_datetime = None

    if reminder_date_time_str is not None:
        reminder_datetime = parse_datetime(reminder_date_time_str)

    todo_api.create_task(name, task_list, reminder_datetime)


def newl(args):
    todo_api.create_list(args.list_name)


def try_parse_as_int(input_str: str):
    try:
        return int(input_str)
    except ValueError:
        return input_str


def complete(args):
    task_list, name = parse_task_path(args.task_name)
    todo_api.complete_task(task_list, try_parse_as_int(name))


def rm(args):
    task_list, name = parse_task_path(args.task_name)
    todo_api.remove_task(task_list, try_parse_as_int(name))


class ArgumentParser(argparse.ArgumentParser):
    class OnExit(Exception):
        pass

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr)
        raise self.OnExit()

    def print_help(self, file=None):
        print_help()


helptext_task_name = """
        Specify the task to complete.
        Can be one of the following:
        task_name
        list_name/task_name
        task_number
        list_name/task_number

        'task_number' is the number displayed when providing the argument '-n'
        """


def setup_parser():
    parser = ArgumentParser(description="Command line interface for Microsoft ToDo")

    parser.set_defaults(func=None)
    subparsers = parser.add_subparsers(help="Command to execute")

    def parser_ls():
        subparser = subparsers.add_parser("ls", help="Display all lists")
        subparser.set_defaults(func=ls)

    def parser_lst():
        subparser = subparsers.add_parser("lst", help="Display tasks from a list")
        subparser.add_argument(
            "list_name",
            nargs="?",
            default="Tasks",
            help="This optional argument specifies the list from which the tasks are displayed."
            "If this parameter is omitted, \
                                    all tasks from the default task list will be displayed",
        )
        subparser.set_defaults(func=lst)

    def parser_new():
        subparser = subparsers.add_parser("new", help="Add a new task")
        subparser.add_argument("task_name", help=helptext_task_name)
        subparser.add_argument("-r", "--reminder")
        subparser.set_defaults(func=new)

    def parser_newl():
        subparser = subparsers.add_parser("newl", help="Add a new list")
        subparser.add_argument("list_name", help="Name of the list to create")
        subparser.set_defaults(func=newl)

    def parser_complete():
        subparser = subparsers.add_parser("complete", help="Complete a Task")
        subparser.add_argument("task_name", help=helptext_task_name)
        subparser.set_defaults(func=complete)

    def parser_rm():
        subparser = subparsers.add_parser("rm", help="Remove a Task")
        subparser.add_argument("task_name", help=helptext_task_name)
        subparser.set_defaults(func=rm)

    parser_lst()
    parser_ls()
    parser_new()
    parser_newl()
    parser_complete()
    parser_rm()

    parser.add_argument(
        "-n",
        "--display_linenums",
        action="store_true",
        default=False,
        help="Display line numbers for the results",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        default=False,
        help="Interactive mode. \
                            Don't exit the application after invoking a command, ask for follow up commands instead.",
    )

    return parser


def main():
    try:
        parser = setup_parser()

        isInteractive = False

        if "-i" in sys.argv or "--interactive" in sys.argv:
            isInteractive = True

        while True:
            try:
                namespace, args = parser.parse_known_args()
                parser.parse_args(args, namespace)

                if namespace.func is not None:
                    namespace.func(namespace)
                else:
                    # No argument was provided
                    parser.print_usage()

            except argparse.ArgumentError:
                pass
            except ArgumentParser.OnExit:
                pass
            except todo_api.TaskNotFoundByName as e:
                eprint(e.message)
            except todo_api.ListNotFound as e:
                eprint(e.message)
            except todo_api.TaskNotFoundByIndex as e:
                eprint(e.message)
            except InvalidTaskPath as e:
                eprint(e.message)
            except TimeExpressionNotRecognized as e:
                eprint(e.message)
            except ErrorParsingTime as e:
                eprint(e.message)
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                pass

            if isInteractive:
                arg = input("\nInput command: ")
                args = shlex.split(arg)
                sys.argv = sys.argv[:1]
                sys.argv += args
            else:
                exit()
    except KeyboardInterrupt:
        print("\n")
        exit(0)


if __name__ == "__main__":
    sys.argv.append("-i")
    main()



from todocli.graphapi.odata_system_query import ODataSystemQuery

base_api_url = "https://graph.microsoft.com/v1.0/me/todo"


def _lists():
    return "{}/lists".format(base_api_url)


def _list(todo_task_list_id):
    return "{}/lists/{}".format(base_api_url, todo_task_list_id)


def _tasks(todo_task_list_id):
    return "{}/lists/{}/tasks".format(base_api_url, todo_task_list_id)


def _task(todo_task_list_id, task_id):
    return "{}/lists/{}/tasks/{}".format(base_api_url, todo_task_list_id, task_id)


def new_list():
    return _lists()


def modify_list(todo_task_list_id):
    return _list(todo_task_list_id)


def all_lists():
    return _lists()


def query_list_id_by_name(list_name):
    return (
        _lists() + ODataSystemQuery().filter_startsWith("displayName", list_name).get()
    )


def new_task(todo_task_list_id):
    return _tasks(todo_task_list_id)


def modify_task(todo_task_list_id, task_id):
    return _task(todo_task_list_id, task_id)


def query_completed_tasks(todo_task_list_id, num_tasks: int):
    return (
        _tasks(todo_task_list_id)
        + ODataSystemQuery().filter_ne("status", "completed").top(num_tasks).get()
    )


def query_task_by_name(todo_task_list_id, task_name: str):
    return (
        _tasks(todo_task_list_id)
        + ODataSystemQuery().filter_eq("title", task_name).get()
    )


def delete_task(todo_task_list_id, task_id):
    return _task(todo_task_list_id, task_id)



help_msg = """
NAME
    todocli - Command line client for Microsoft ToDo 

SYNOPSIS
    todocli [options] COMMAND ...  

    'COMMAND' can be one of the following values:
        ls                  Display all lists  

        lst <list_name>     Display all tasks from list
            list_name       Name of the list

        new <task> [-r time]
                            Create a new task
            task            Task to create. See 'Specifying a task' for details.
            -r time         Set a reminder. See 'Specifying time' for details.              

        newl <list_name>    Create a new list
            list_name       Name of the list

        complete <task>     Set task status to completed
            task            Task to complete. See 'Specifying a task' for details.

        rm <task>           Remove a task
            task            Task to remove. See 'Specifying a task' for details.

OPTIONS
    -h, --help
        Display a usage message.

    -i, --interactive
        Interactive mode. 
        Don't exit after invoking a command, but ask for follow up commands instead.

    -n, --display_linenums
        Display a line number for all lines which are output.

Specifying a task:
    For commands which take 'task' as a parameter, 'task' can be one of the following:

    task_name
    list_name/task_name
    task_number
    list_name/task_number

    If 'list_name' is omitted, the default task list will be used. 
    'task_number' is the position displayed when specifying option '-n'. 

Specifying time:
    For options which take 'time' as a parameter, 'time' can be one of the following:

    {n}h
        Current time + n hours. 
        e.g. 1h, 12h. 
        Max: 99h

    morning
        Today at 07:00 AM if current time < 07:00 AM, otherwise tomorrow

    evening
        Today at 06:00 PM if current time < 06:00 PM, otherwise tomorrow

    tomorrow
        Tomorrow at 07:00 AM

    {hour}:{minute}
        Today at {hour}:{minute} if current time < {hour}:{minute}, otherwise tomorrow 
        e.g. 9:30, 09:30, 17:15

    {hour}:{minute} am|pm 
        Today at {hour}:{minute} am|pm  if current time < {hour}:{minute} am|pm, otherwise tomorrow
        e.g. 9:30 am, 12:00 am, 10:15 pm

    {day}.{month}. {hour}:{minute}
        The given day at {hour}:{minute}
        e.g. 24.12. 12:00
        e.g. 7.4.   9:15

    {day}.{month}.{year}
        The given day at 7:00 am
        e.g. 22.12.2020
        e.g. 01.01.21
        """



from enum import Enum


class ODataSystemQuery:
    class Format(Enum):
        json = "json"
        atom = "atom"
        xml = "xml"

    class SortOrder(Enum):
        Ascending = "asc"
        Descending = "desc"

    def add_to_query(self, expr):
        if len(self.query_expr) == 1:
            self.query_expr += expr
        else:
            self.query_expr += "&" + expr

    def __init__(self):
        self.query_expr = "?"

    def count(self):
        self.add_to_query("$count=true")
        return self

    def expand(self, expr):
        self.add_to_query("$expand={}".format(expr))
        return self

    def filter_startsWith(self, property_name, expr):
        self.filter("startsWith({},'{}')".format(property_name, expr))
        return self

    def filter_endsWith(self, property_name, expr):
        self.filter("endsWith({},'{}')".format(property_name, expr))
        return self

    def _filter_compare(self, property_name, comparison, expr):
        self.filter("{} {} '{}'".format(property_name, comparison, expr))
        return self

    def filter_eq(self, property_name, expr):
        self._filter_compare(property_name, "eq", expr)
        return self

    def filter_ne(self, property_name, expr):
        self._filter_compare(property_name, "ne", expr)
        return self

    def filter_gt(self, property_name, expr):
        self._filter_compare(property_name, "gt", expr)
        return self

    def filter_ge(self, property_name, expr):
        self._filter_compare(property_name, "ge", expr)
        return self

    def filter_lt(self, property_name, expr):
        self._filter_compare(property_name, "lt", expr)
        return self

    def filter_le(self, property_name, expr):
        self._filter_compare(property_name, "le", expr)
        return self

    def filter(self, expr):
        self.add_to_query("$filter={}".format(expr))
        return self

    def format(self, ftmat: Format):
        self.add_to_query("$format={}".format(ftmat.value))
        return self

    def orderBy(self, expr, sort_order: SortOrder = SortOrder.Ascending):
        self.add_to_query("$orderby={} {}".format(expr, sort_order.value))
        return self

    def search(self, expr):
        raise Exception("Not implemented")

    def select(self, expr):
        raise Exception("Not implemented")

    def skip(self, expr):
        raise Exception("Not implemented")

    def top(self, num_elements):
        self.add_to_query("$top={}".format(num_elements))
        return self

    def get(self):
        if self.query_expr == "?":
            return ""
        return self.query_expr



"""
For implementation details, refer to this source:
https://docs.microsoft.com/en-us/graph/api/resources/todo-overview?view=graph-rest-1.0
"""
import os
import pickle

from datetime import datetime
from typing import Union

from todocli.models.todolist import TodoList
from todocli.models.task import Task

from todocli.graphapi import api_urls
from todocli.graphapi.oauth import config_dir
from todocli.graphapi.rest_request import (
    RestRequestGet,
    RestRequestPost,
    RestRequestPatch,
    RestRequestDelete,
    RestRequestWithBody,
)
from todocli.utils.datetime_util import datetime_to_api_timestamp

list_ids_cached = {}


class ListNotFound(Exception):
    def __init__(self, list_name):
        self.message = "List with name '{}' could not be found".format(list_name)
        super(ListNotFound, self).__init__(self.message)


class TaskNotFoundByName(Exception):
    def __init__(self, task_name, list_name):
        self.message = "Task with name '{}' could not be found in list '{}'".format(
            task_name, list_name
        )
        super(TaskNotFoundByName, self).__init__(self.message)


class TaskNotFoundByIndex(Exception):
    def __init__(self, task_index, list_name):
        self.message = "Task with index '{}' could not be found in list '{}'".format(
            task_index, list_name
        )
        super(TaskNotFoundByIndex, self).__init__(self.message)


class _RestRequestTask:
    def __init__(self):
        self.request = None

    def set_completed(self):
        self.request["completedDateTime"] = datetime_to_api_timestamp(datetime.now())
        self.set_status(Task.Status.Completed)

    def set_status(self, status: Task.Status):
        self.request["status"] = status.value

    def set_importance(self, importance: Task.Importance):
        self.request["importance"] = importance.value

    def set_title(self, title: str):
        self.request["title"] = title

    def set_reminder(self, reminder_datetime):
        self.request["isReminderOn"] = True
        self.request["reminderDateTime"] = datetime_to_api_timestamp(reminder_datetime)

    def execute(self):
        return self.request.execute()


class RestRequestTaskModify(_RestRequestTask):
    def __init__(self, list_name, task_name):
        super().__init__()

        url = api_urls.modify_task(
            get_list_id_by_name(list_name), get_task_id(list_name, task_name)
        )
        self.request = RestRequestPatch(url)


class RestRequestTaskNew(_RestRequestTask):
    def __init__(self, list_name, task_name):
        super().__init__()

        url = api_urls.new_task(get_list_id_by_name(list_name))
        self.request = RestRequestPost(url)
        self.set_title(task_name)

    def _get_request(self) -> RestRequestWithBody:
        return self.request


def query_list_id_by_name(list_name):
    url = api_urls.query_list_id_by_name(list_name)
    result = RestRequestGet(url).execute()

    try:
        return result[0]["id"]
    except IndexError:
        raise ListNotFound(list_name)


def get_list_id_by_name(list_name: str):
    if list_name not in list_ids_cached:
        list_id = query_list_id_by_name(list_name)
        list_ids_cached[list_name] = list_id
        return list_id
    else:
        return list_ids_cached[list_name]


def query_tasks(list_name: str, num_tasks: int = 100):
    query_url = api_urls.query_completed_tasks(
        get_list_id_by_name(list_name), num_tasks
    )
    result = RestRequestGet(query_url).execute()
    return [Task(x) for x in result]


def query_task(list_name: str, task_name: str):
    query_url = api_urls.query_task_by_name(get_list_id_by_name(list_name), task_name)
    result = RestRequestGet(query_url).execute()
    return [Task(x) for x in result]


def create_list(title: str):
    request = RestRequestPost(api_urls.new_list())
    request["displayName"] = title
    return request.execute()


def rename_list(old_list_title: str, new_list_title: str):
    request = RestRequestPatch(
        api_urls.modify_list(get_list_id_by_name(old_list_title))
    )
    request["title"] = new_list_title
    return request.execute()


def create_task(task_name: str, list_name: str, reminder_datetime: datetime = None):
    request = RestRequestTaskNew(list_name, task_name)

    if reminder_datetime is not None:
        request.set_reminder(reminder_datetime)

    return request.execute()


def query_lists():
    result = RestRequestGet(api_urls.all_lists()).execute()
    lists = [TodoList(x) for x in result]
    return lists


def get_task_id_by_name(list_name: str, task_name: str):
    try:
        return query_task(list_name, task_name)[0].id
    except IndexError:
        raise TaskNotFoundByName(task_name, list_name)


def get_task_id_by_list_position(list_name: str, task_list_position):
    tasks = query_tasks(list_name, task_list_position + 1)
    try:
        return tasks[task_list_position].id
    except IndexError:
        raise TaskNotFoundByIndex(task_list_position, list_name)


def get_task_id(list_name: str, task_name_or_listpos: Union[str, int]):
    if isinstance(task_name_or_listpos, str):
        return get_task_id_by_name(list_name, task_name_or_listpos)
    elif isinstance(task_name_or_listpos, int):
        return get_task_id_by_list_position(list_name, task_name_or_listpos)
    else:
        raise


def complete_task(list_name: str, task_name: Union[str, int]):
    request = RestRequestTaskModify(list_name, task_name)
    request.set_completed()
    request.execute()


def remove_task(task_list, param):
    task_id = get_task_id(task_list, param)
    url = api_urls.delete_task(task_list, task_id)
    request = RestRequestDelete(url)
    request.execute()



from enum import Enum
from todocli.utils.datetime_util import api_timestamp_to_datetime


class Task:
    class Status(Enum):
        Completed = "completed"
        NotStarted = "notStarted"
        InProgress = "inProgress"
        WaitingOnOthers = "waitingOnOthers"
        Deferred = "deferred"

    class Importance(Enum):
        Low = "low"
        Normal = "normal"
        High = "high"

    def __init__(self, query_result):
        self.title = query_result["title"]
        self.id = query_result["id"]
        self.importance = Task.Importance(query_result["importance"])
        self.status = Task.Status(query_result["status"])
        self.created_datetime = api_timestamp_to_datetime(
            query_result["createdDateTime"]
        )

        if "completedDateTime" in query_result:
            self.completed_datetime = api_timestamp_to_datetime(
                query_result["completedDateTime"]
            )
        else:
            self.completed_datetime = None

        self.is_reminder_on: bool = bool(query_result["isReminderOn"])

        if "reminderDateTime" in query_result:
            self.reminder_datetime = api_timestamp_to_datetime(
                query_result["reminderDateTime"]
            )
        else:
            self.reminder_datetime = None

        self.last_modified_datetime = api_timestamp_to_datetime(
            query_result["lastModifiedDateTime"]
        )

        if "bodyLastModifiedDateTime" in query_result:
            self.body_last_modified_datetime = api_timestamp_to_datetime(
                query_result["bodyLastModifiedDateTime"]
            )
        else:
            self.body_last_modified_datetime = None



import json
import os
import pickle

from todocli.graphapi.oauth import get_oauth_session, config_dir

base_api_url = "https://graph.microsoft.com/v1.0/me/todo"


def parse_contents(response):
    return json.loads(response.content.decode())["value"]


def list_tasks(all_=False, folder=""):
    outlook = get_oauth_session()

    if folder == "":
        if all_:
            o = outlook.get("{}/tasks?top=100".format(base_api_url))
        else:
            o = outlook.get(
                "{}/tasks?$filter=status ne 'completed'&top=100".format(base_api_url)
            )
    else:
        if all_:
            o = outlook.get("{}/lists/{}/tasks?top=100".format(base_api_url, folder))
        else:
            o = outlook.get(
                "{}/lists/{}/tasks?$filter=status ne 'completed'&top=100".format(
                    base_api_url, folder
                )
            )

    return parse_contents(o)


def list_and_update_folders():
    outlook = get_oauth_session()
    o = outlook.get("{}/lists?top=20".format(base_api_url))
    contents = parse_contents(o)

    # Cache folders
    name2id = {}
    id2name = {}

    folders = parse_contents(o)
    for f in folders:
        name2id[f["displayName"]] = f["id"]
        id2name[f["id"]] = f["displayName"]

    with open(os.path.join(config_dir, "folder_name2id.pkl"), "wb") as f:
        pickle.dump(name2id, f)
    with open(os.path.join(config_dir, "folder_id2name.pkl"), "wb") as f:
        pickle.dump(id2name, f)

    return contents


def create_folder(name):
    """Create folder with name `name`"""
    outlook = get_oauth_session()

    # Fill request body
    request_body = {"name": name}

    o = outlook.post("{}/lists".format(base_api_url), json=request_body)

    return o.ok


def delete_folder(folder_id):
    """Delete folder with id `folder_id`"""
    outlook = get_oauth_session()
    o = outlook.delete("{}/lists/{}".format(base_api_url, folder_id))
    return o.ok


def create_task(text, folder=None):
    """Create task with subject `text`"""
    outlook = get_oauth_session()

    # Fill request body
    request_body = {"title": text}

    if folder is None:
        o = outlook.post("{}/tasks".format(base_api_url), json=request_body)
    else:
        o = outlook.post(
            "{}/lists/{}/tasks".format(base_api_url, folder), json=request_body
        )

    return o.ok


def delete_task(task_id):
    outlook = get_oauth_session()

    o = outlook.delete("{}/tasks/{}".format(base_api_url, task_id))
    return o.ok


def complete_task(list_id, task_id):
    outlook = get_oauth_session()

    o = outlook.patch(
        "{}/lists/{}/tasks/{}/".format(base_api_url, list_id, task_id),
        json={"status": "completed"},
    )
    return o.ok



from setuptools import setup, find_packages

setup(
    name="tod0",
    version="0.6.5",
    author="kiblee",
    author_email="kiblee@pm.me",
    packages=find_packages(),
    url="https://github.com/kiblee/tod0",
    license="LICENSE",
    description="A Terminal Client for Microsoft To-Do.",
    install_requires=[
        "prompt-toolkit",
        "pyyaml",
        "requests",
        "requests_oauthlib",
    ],
    include_package_data=True,
    entry_points="""
        [console_scripts]
        tod0=todocli.interface:run
        todocli=todocli.cli:main
    """,
    python_requires=">=3.6",
)



import unittest
import todocli

from unittest.mock import Mock, patch
from datetime import datetime, timedelta

from todocli.utils.datetime_util import (
    parse_datetime,
    add_day_if_past,
    ErrorParsingTime,
    TimeExpressionNotRecognized,
)


class TestDatetimeParser(unittest.TestCase):
    @patch.object(todocli.utils.datetime_util, "datetime", Mock(wraps=datetime))
    def test_add_day_if_datetime_is_in_past(self):
        dt_now = datetime(2020, 1, 1, 9, 0)
        todocli.utils.datetime_util.datetime.now.return_value = dt_now
        dt = dt_now - timedelta(minutes=1)
        dt = add_day_if_past(dt)
        self.assertGreater(dt.day, dt_now.day)

        dt = dt_now + timedelta(hours=1)
        dt = add_day_if_past(dt)
        self.assertEqual(dt.day, dt_now.day)

    def test_am_pm_time1(self):
        input_str = "07:00 pm"
        dt = parse_datetime(input_str)
        dt_expected = datetime.now().replace(hour=19, minute=0, second=0, microsecond=0)
        dt_expected = add_day_if_past(dt_expected)
        self.assertEqual(dt, dt_expected)

    def test_am_pm_time2(self):
        input_str = "7:00 pm"
        dt = parse_datetime(input_str)
        dt_expected = datetime.now().replace(hour=19, minute=0, second=0, microsecond=0)
        dt_expected = add_day_if_past(dt_expected)
        self.assertEqual(dt, dt_expected)

    def test_am_pm_time3(self):
        input_str = "12:00 am"
        dt = parse_datetime(input_str)
        dt_expected = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        dt_expected = add_day_if_past(dt_expected)
        self.assertEqual(dt, dt_expected)

    @patch.object(todocli.utils.datetime_util, "datetime", Mock(wraps=datetime))
    def test_time_strings(self):
        # (user input, simulated 'now' time, expected output)
        times = [
            ("07:00", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 7, 0)),
            ("7:00", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 7, 0)),
            ("0:00", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 0, 0)),
            ("17:30", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 1, 17, 30)),
            ("6:30 am", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 6, 30)),
            ("6:30 pm", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 1, 18, 30)),
            ("06:30 am", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 6, 30)),
            ("06:30 pm", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 1, 18, 30)),
            ("00:00 am", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 0, 0)),
            ("12:00 am", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 0, 0)),
            ("24.06. 12:00", datetime(2020, 6, 24, 9, 0), datetime(2020, 6, 24, 12, 0)),
            ("24.06. 6:00", datetime(2020, 6, 23, 9, 0), datetime(2020, 6, 24, 6, 0)),
            ("24.6. 6:00", datetime(2020, 6, 23, 9, 0), datetime(2020, 6, 24, 6, 0)),
            ("24.6. 06:00", datetime(2020, 6, 23, 9, 0), datetime(2020, 6, 24, 6, 0)),
            ("24.6. 12:00", datetime(2020, 6, 24, 9, 0), datetime(2020, 6, 24, 12, 0)),
            ("6.6. 06:06", datetime(2020, 6, 5, 9, 0), datetime(2020, 6, 6, 6, 6)),
            ("6.6. 6:06", datetime(2020, 6, 5, 9, 0), datetime(2020, 6, 6, 6, 6)),
            # "04/27 12:00 am",
            # "04/27 2:00 am",
            # "04/27 3:00 pm",
            # "04/27 10:00 pm",
            # "4/27 10:00 pm",
            # "4/27 09:00 pm",
            # "4/27 9:00 pm",
            # "4/4 9:00 pm",
            # "4/4 09:00 pm",
            # "4/4 12:00 pm",
            ("morning", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 7, 0)),
            ("evening", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 1, 18, 0)),
            ("tomorrow", datetime(2020, 1, 1, 9, 0), datetime(2020, 1, 2, 7, 0)),
        ]

        for user_input, simulated_now_time, expected_output in times:
            print(f"testing time: {user_input}")
            todocli.utils.datetime_util.datetime.now.return_value = simulated_now_time
            return_val = parse_datetime(user_input)
            self.assertEqual(return_val, expected_output)

    def test_invalid_time(self):
        invalid_times = [
            "24:00",
            "25:00",
            "25123:00",
            "0:12345",
            "12:30 sam",
            "12:30 pom",
            "abfdsa",
        ]

        for time in invalid_times:
            print(f"testing time: {time}")
            try:
                self.assertRaises(Exception, parse_datetime(time))
            except ErrorParsingTime:
                pass
            except TimeExpressionNotRecognized:
                pass
            else:
                raise


if __name__ == "__main__":
    unittest.main()
